-- ============================================================= --
-- FOG CONTROL MOD
-- ============================================================= --
FogControl = {}
FogControl.name = g_currentModName
FogControl.path = g_currentModDirectory

FogControl.SETTINGS = {}
FogControl.CONTROLS = {}

addModEventListener(FogControl)

FogControl.menuItems = {
    'fogEnabled',
    'groundFogDensity',
    'heightFogDensity',
    'fogCoverage',
    'fogStartTime',
    'fogEndTime',
    'groundFogExtraHeight',
    'groundFogMinValleyDepth',
    'heightFogMaxHeight',
    'targetValue',
    'lastMieScale',
    'nightFactor',
    'dayFactor',
}

-- Default settings
FogControl.fogEnabled = true
FogControl.groundFogDensity = 0.005
FogControl.heightFogDensity = 0.6
FogControl.fogCoverage = 0.015
FogControl.fogStartTime = 240
FogControl.fogEndTime = 600
FogControl.groundFogExtraHeight = 7
FogControl.groundFogMinValleyDepth = 1.5
FogControl.heightFogMaxHeight = 550
FogControl.targetValue = 100
FogControl.lastMieScale = 100
FogControl.nightFactor = 0.5
FogControl.dayFactor = 0.5
FogControl.lastCheckedDay = -1
FogControl.lastUpdateTime = -1

function FogControl:debugPrint(str)
    if g_currentMission and g_currentMission.isMasterUser then
        print("[FogControl] " .. str)
    end
end

function FogControl:updateFogSettings()
    if not g_currentMission or not g_currentMission.environment or not g_currentMission.environment.weather then
        FogControl:debugPrint("Environment or weather not initialized")
        return
    end

    local fogUpdater = g_currentMission.environment.weather.fogUpdater
    local fogSettings = FogSettings.new()
    local currentDayTime = g_currentMission.environment.weather.owner.dayTime / 60000 -- Convert to minutes
    local weather = g_currentMission.environment.weather
    local currentWeatherType = weather:getCurrentWeatherType()

    -- Suitable weather types for fog
    local suitableWeatherTypes = {WeatherType.FOG, WeatherType.RAIN, WeatherType.CLOUDY}
    local isSuitableWeather = false
    for _, fogType in ipairs(suitableWeatherTypes) do
        if currentWeatherType == fogType then
            isSuitableWeather = true
            break
        end
    end

    if not FogControl.fogEnabled or
       currentDayTime < FogControl.fogStartTime or
       currentDayTime > FogControl.fogEndTime or
       not isSuitableWeather then
        -- Disable fog if outside time window, disabled, or unsuitable weather
        fogSettings.groundFogGroundLevelDensity = 0
        fogSettings.heightFogGroundLevelDensity = 0
        fogSettings.groundFogCoverageEdge0 = 0
        fogSettings.groundFogCoverageEdge1 = 0
        fogSettings.groundFogStartDayTimeMinutes = 0
        fogSettings.groundFogEndDayTimeMinutes = 0
        fogSettings.heightFogMaxHeight = 0
        fogUpdater.visibilityAlpha = 0
        local reason = not FogControl.fogEnabled and "Fog disabled" or
                       (currentDayTime < FogControl.fogStartTime or currentDayTime > FogControl.fogEndTime) and
                       string.format("Outside fog time window (%d-%d)", FogControl.fogStartTime, FogControl.fogEndTime) or
                       "Unsuitable weather: " .. tostring(currentWeatherType)
        FogControl:debugPrint("Fog disabled: " .. reason)
    else
        -- Apply user settings
        fogSettings.groundFogGroundLevelDensity = FogControl.groundFogDensity
        fogSettings.heightFogGroundLevelDensity = FogControl.heightFogDensity
        fogSettings.groundFogCoverageEdge0 = FogControl.fogCoverage
        fogSettings.groundFogCoverageEdge1 = math.max(FogControl.fogCoverage, FogControl.fogCoverage + 0.01)
        fogSettings.groundFogExtraHeight = FogControl.groundFogExtraHeight
        fogSettings.groundFogMinValleyDepth = FogControl.groundFogMinValleyDepth
        fogSettings.heightFogMaxHeight = FogControl.heightFogMaxHeight
        fogSettings.groundFogStartDayTimeMinutes = FogControl.fogStartTime
        fogSettings.groundFogEndDayTimeMinutes = FogControl.fogEndTime

        -- Validate start and end times
        if fogSettings.groundFogStartDayTimeMinutes >= fogSettings.groundFogEndDayTimeMinutes then
            fogSettings.groundFogEndDayTimeMinutes = fogSettings.groundFogStartDayTimeMinutes + 60
            FogControl.fogEndTime = fogSettings.groundFogEndDayTimeMinutes
            FogControl:debugPrint("Corrected fogEndTime to be after fogStartTime")
        end

        fogUpdater.visibilityAlpha = 1
        FogControl:debugPrint(string.format("Fog enabled with user settings: groundFogDensity=%.4f, heightFogDensity=%.4f, fogCoverage=%.4f, extraHeight=%.1f, minValleyDepth=%.1f, maxHeight=%.1f, weather=%s",
            fogSettings.groundFogGroundLevelDensity, fogSettings.heightFogGroundLevelDensity, fogSettings.groundFogCoverageEdge0,
            fogSettings.groundFogExtraHeight, fogSettings.groundFogMinValleyDepth, fogSettings.heightFogMaxHeight, tostring(currentWeatherType)))
    end

    -- Apply fog state parameters
    local fogState = {
        targetValue = FogControl.targetValue,
        lastMieScale = FogControl.lastMieScale,
        alpha = fogUpdater.visibilityAlpha,
        duration = 30000, -- 30-second transition
        nightFactor = FogControl.nightFactor,
        dayFactor = FogControl.dayFactor,
    }
    fogUpdater:setTargetFog(fogSettings, fogState.duration, fogState)
    fogUpdater.isDirty = true
    FogControl.lastUpdateTime = currentDayTime
end

function FogControl:update(dt)
    if not g_currentMission or not g_currentMission:getIsServer() then
        return
    end

    if not FogControl.initialised then
        FogControl:updateFogSettings()
        FogControl.initialised = true
        return
    end

    local environment = g_currentMission.environment
    if not environment or not environment.weather then
        return
    end

    local currentDay = environment.weather.owner.currentMonotonicDay
    local currentDayTime = environment.dayTime / 60000 -- Convert to minutes

    -- Reset lastCheckedDay when a new day starts
    if currentDay ~= FogControl.lastCheckedDay then
        FogControl.lastCheckedDay = currentDay
        FogControl:debugPrint("New day detected, resetting fog check")
    end

    -- Periodically check fog state (every 5 minutes)
    if FogControl.lastUpdateTime == -1 or math.abs(currentDayTime - FogControl.lastUpdateTime) >= 5 then
        FogControl:debugPrint("Periodic fog check at " .. currentDayTime .. " minutes")
        FogControl:updateFogSettings()
    end
end

function FogControl:loadMap(name)
    FogControl:readSettings()
    addConsoleCommand("gsFogControl", "Control fog settings", "consoleCommandFogControl", FogControl)
    FogControl:debugPrint("Map loaded, applying initial fog settings")
    FogControl:updateFogSettings()
end

function FogControl:onDayChanged()
    FogControl:debugPrint("New day detected, reapplying fog settings")
    FogControl:updateFogSettings()
end

function FogControl:onWeatherObjectChanged()
    FogControl:debugPrint("Weather changed, reapplying fog settings")
    FogControl:updateFogSettings()
end

-- Console command
function FogControl:consoleCommandFogControl(param1, param2)
    if param1 == "enable" then
        local value = param2:lower() == "true"
        FogControl.fogEnabled = value
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Fog enabled: %s", tostring(value))
    elseif param1 == "groundDensity" then
        local value = tonumber(param2) or FogControl.groundFogDensity
        FogControl.groundFogDensity = math.clamp(value, 0.0, 1.0)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Ground fog density: %.4f", FogControl.groundFogDensity)
    elseif param1 == "heightDensity" then
        local value = tonumber(param2) or FogControl.heightFogDensity
        FogControl.heightFogDensity = math.clamp(value, 0.0, 1.0)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Height fog density: %.4f", FogControl.heightFogDensity)
    elseif param1 == "coverage" then
        local value = tonumber(param2) or FogControl.fogCoverage
        FogControl.fogCoverage = math.clamp(value, 0.0, 1.0)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Fog coverage: %.4f", FogControl.fogCoverage)
    elseif param1 == "startTime" then
        local value = tonumber(param2) or FogControl.fogStartTime
        FogControl.fogStartTime = math.clamp(value, 0, 1440)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Fog start time: %d minutes", FogControl.fogStartTime)
    elseif param1 == "endTime" then
        local value = tonumber(param2) or FogControl.fogEndTime
        FogControl.fogEndTime = math.clamp(value, 0, 1440)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Fog end time: %d minutes", FogControl.fogEndTime)
    elseif param1 == "extraHeight" then
        local value = tonumber(param2) or FogControl.groundFogExtraHeight
        FogControl.groundFogExtraHeight = math.clamp(value, 0, 50)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Ground fog extra height: %.1f", FogControl.groundFogExtraHeight)
    elseif param1 == "minValleyDepth" then
        local value = tonumber(param2) or FogControl.groundFogMinValleyDepth
        FogControl.groundFogMinValleyDepth = math.clamp(value, 0.5, 20)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Ground fog min valley depth: %.1f", FogControl.groundFogMinValleyDepth)
    elseif param1 == "maxHeight" then
        local value = tonumber(param2) or FogControl.heightFogMaxHeight
        FogControl.heightFogMaxHeight = math.clamp(value, 0, 1500)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Height fog max height: %.1f", FogControl.heightFogMaxHeight)
    elseif param1 == "targetValue" then
        local value = tonumber(param2) or FogControl.targetValue
        FogControl.targetValue = math.clamp(value, 0, 200)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Target value: %.1f", FogControl.targetValue)
    elseif param1 == "lastMieScale" then
        local value = tonumber(param2) or FogControl.lastMieScale
        FogControl.lastMieScale = math.clamp(value, 0, 200)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Last Mie scale: %.1f", FogControl.lastMieScale)
    elseif param1 == "nightFactor" then
        local value = tonumber(param2) or FogControl.nightFactor
        FogControl.nightFactor = math.clamp(value, 0, 0.75)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Night factor: %.2f", FogControl.nightFactor)
    elseif param1 == "dayFactor" then
        local value = tonumber(param2) or FogControl.dayFactor
        FogControl.dayFactor = math.clamp(value, 0, 0.75)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Day factor: %.2f", FogControl.dayFactor)
    end
    return "Usage: gsFogControl [enable|groundDensity|heightDensity|coverage|startTime|endTime|extraHeight|minValleyDepth|maxHeight|targetValue|lastMieScale|nightFactor|dayFactor] <value>"
end

-- Settings management
function FogControl:setValue(id, value)
    FogControl[id] = value
end

function FogControl:getValue(id)
    return FogControl[id]
end

function FogControl:getStateIndex(id, value)
    local value = value or FogControl:getValue(id)
    local setting = FogControl.SETTINGS[id]
    if not setting then
        return 1
    end
    local values = setting.values
    if type(value) == 'number' then
        local index = setting.default
        local initialDiff = math.huge
        for i, v in pairs(values) do
            local currentDiff = math.abs(v - value)
            if currentDiff < initialDiff then
                initialDiff = currentDiff
                index = i
            end
        end
        return index
    else
        for i, v in pairs(values) do
            if value == v then
                return i
            end
        end
    end
    return setting.default
end

function FogControl:writeSettings()
    local key = "fogControlSettings"
    local userSettingsFile = Utils.getFilename("modSettings/FogControl.xml", getUserProfileAppPath())
    local xmlFile = createXMLFile("settings", userSettingsFile, key)

    if xmlFile ~= 0 then
        for _, id in pairs(FogControl.menuItems) do
            local value = FogControl:getValue(id)
            local xmlValueKey = "fogControlSettings." .. id .. "#value"
            if type(value) == 'number' then
                setXMLFloat(xmlFile, xmlValueKey, value)
            elseif type(value) == 'boolean' then
                setXMLBool(xmlFile, xmlValueKey, value)
            end
        end
        saveXMLFile(xmlFile)
        delete(xmlFile)
    end
end

function FogControl:readSettings()
    local userSettingsFile = Utils.getFilename("modSettings/FogControl.xml", getUserProfileAppPath())
    if not fileExists(userSettingsFile) then
        FogControl:writeSettings()
        return
    end

    local xmlFile = loadXMLFile("fogControlSettings", userSettingsFile)
    if xmlFile ~= 0 then
        for _, id in pairs(FogControl.menuItems) do
            local setting = FogControl.SETTINGS[id]
            if setting then
                local xmlValueKey = "fogControlSettings." .. id .. "#value"
                if hasXMLProperty(xmlFile, xmlValueKey) then
                    local value = FogControl:getValue(id)
                    if type(value) == 'number' then
                        value = getXMLFloat(xmlFile, xmlValueKey) or value
                    elseif type(value) == 'boolean' then
                        value = getXMLBool(xmlFile, xmlValueKey) or value
                    end
                    FogControl:setValue(id, value)
                end
            end
        end
        delete(xmlFile)
    end
end

-- Menu integration
FogControl.SETTINGS.fogEnabled = {
    default = 2,
    values = {false, true},
    strings = {g_i18n:getText("ui_off"), g_i18n:getText("ui_on")}
}

-- Generate percentage values for density and coverage (0% to 100% with 0.1% step)
local percentValues = {}
local percentStrings = {}
for i = 0, 1000 do
    percentValues[i + 1] = i / 1000 -- 0.000 to 1.000 with step 0.001
    percentStrings[i + 1] = string.format("%.1f%%", i / 10)
end

FogControl.SETTINGS.groundFogDensity = {
    default = 5, -- 0.5% (0.005)
    values = percentValues,
    strings = percentStrings
}

FogControl.SETTINGS.heightFogDensity = {
    default = 600, -- 60% (0.6)
    values = percentValues,
    strings = percentStrings
}

FogControl.SETTINGS.fogCoverage = {
    default = 15, -- 1.5% (0.015)
    values = percentValues,
    strings = percentStrings
}

FogControl.SETTINGS.fogStartTime = {
    default = 2,
    values = {0, 240, 300, 360, 420},
    strings = {"00:00", "04:00", "05:00", "06:00", "07:00"}
}

FogControl.SETTINGS.fogEndTime = {
    default = 2,
    values = {480, 540, 600, 660, 720},
    strings = {"08:00", "09:00", "10:00", "11:00", "12:00"}
}

FogControl.SETTINGS.groundFogExtraHeight = {
    default = 7,
    values = {0, 5, 7, 10, 20, 50},
    strings = {"0", "5", "7", "10", "20", "50"}
}

FogControl.SETTINGS.groundFogMinValleyDepth = {
    default = 2,
    values = {0.5, 1, 1.5, 2, 5, 10, 20},
    strings = {"0.5", "1", "1.5", "2", "5", "10", "20"}
}

FogControl.SETTINGS.heightFogMaxHeight = {
    default = 2,
    values = {0, 100, 550, 1000, 1500},
    strings = {"0", "100", "550", "1000", "1500"}
}

FogControl.SETTINGS.targetValue = {
    default = 5,
    values = {0, 50, 100, 150, 200},
    strings = {"0", "50", "100", "150", "200"}
}

FogControl.SETTINGS.lastMieScale = {
    default = 5,
    values = {0, 50, 100, 150, 200},
    strings = {"0", "50", "100", "150", "200"}
}

FogControl.SETTINGS.nightFactor = {
    default = 3,
    values = {0, 0.25, 0.5, 0.75},
    strings = {"0", "0.25", "0.5", "0.75"}
}

FogControl.SETTINGS.dayFactor = {
    default = 3,
    values = {0, 0.25, 0.5, 0.75},
    strings = {"0", "0.25", "0.5", "0.75"}
}

function FogControl:addMenuOption(id)
    local inGameMenu = g_gui.screenControllers[InGameMenu]
    local settingsPage = inGameMenu.pageSettings
    local settingsLayout = settingsPage.generalSettingsLayout

    local original = #FogControl.SETTINGS[id].values == 2 and settingsPage.checkWoodHarvesterAutoCutBox or settingsPage.multiVolumeVoiceBox
    local options = FogControl.SETTINGS[id].strings

    local menuOptionBox = original:clone(settingsLayout)
    menuOptionBox.id = id .. "box"

    local menuOption = menuOptionBox.elements[1]
    menuOption.id = id
    menuOption.target = FogControl
    menuOption:setCallback("onClickCallback", "onMenuOptionChanged")
    menuOption:setDisabled(false)

    local toolTip = menuOption.elements[1]
    toolTip:setText(g_i18n:getText("tooltip_fogcontrol_" .. id))

    local setting = menuOptionBox.elements[2]
    setting:setText(g_i18n:getText("setting_fogcontrol_" .. id))

    menuOption:setTexts({unpack(options)})
    menuOption:setState(FogControl:getStateIndex(id))

    FogControl.CONTROLS[id] = menuOption

    local function updateFocusIds(element)
        if not element then return end
        element.focusId = FocusManager:serveAutoFocusId()
        for _, child in pairs(element.elements) do
            updateFocusIds(child)
        end
    end
    updateFocusIds(menuOptionBox)
    table.insert(settingsPage.controlsList, menuOptionBox)
end

function FogControl:onMenuOptionChanged(state, menuOption)
    local id = menuOption.id
    local value = FogControl.SETTINGS[id].values[state]
    FogControl:setValue(id, value)
    FogControl:writeSettings()
    FogControl:updateFogSettings()
end

-- Add menu section and options
local inGameMenu = g_gui.screenControllers[InGameMenu]
local settingsPage = inGameMenu.pageSettings
local settingsLayout = settingsPage.generalSettingsLayout

local sectionTitle
for _, elem in ipairs(settingsLayout.elements) do
    if elem.name == "sectionHeader" then
        sectionTitle = elem:clone(settingsLayout)
        break
    end
end
if sectionTitle then
    sectionTitle:setText(g_i18n:getText("menu_FogControl_TITLE"))
else
    local title = TextElement.new()
    title:applyProfile("fs25_settingsSectionHeader", true)
    title:setText(g_i18n:getText("menu_FogControl_TITLE"))
    title.name = "sectionHeader"
    settingsLayout:addElement(title)
    sectionTitle = title
end

sectionTitle.focusId = FocusManager:serveAutoFocusId()
table.insert(settingsPage.controlsList, sectionTitle)
FogControl.CONTROLS[sectionTitle.name] = sectionTitle

for _, id in pairs(FogControl.menuItems) do
    FogControl:addMenuOption(id)
end
settingsLayout:invalidateLayout()

-- Handle focus for custom controls
FocusManager.setGui = Utils.appendedFunction(FocusManager.setGui, function(_, gui)
    if gui == "ingameMenuSettings" then
        for _, control in pairs(FogControl.CONTROLS) do
            if not control.focusId or not FocusManager.currentFocusData.idToElementMapping[control.focusId] then
                if not FocusManager:loadElementFromCustomValues(control, nil, nil, false, false) then
                    Logging.warning("Could not register control %s with the focus manager", control.id or control.name or control.focusId)
                end
            end
        end
        local settingsPage = g_gui.screenControllers[InGameMenu].pageSettings
        settingsPage.generalSettingsLayout:invalidateLayout()
    end
end)

InGameMenuSettingsFrame.onFrameOpen = Utils.appendedFunction(InGameMenuSettingsFrame.onFrameOpen, function()
    local isAdmin = g_currentMission:getIsServer() or g_currentMission.isMasterUser
    for _, id in pairs(FogControl.menuItems) do
        local menuOption = FogControl.CONTROLS[id]
        menuOption:setState(FogControl:getStateIndex(id))
        menuOption:setDisabled(not isAdmin)
    end
end)

-- Register for events
Environment.onDayChanged = Utils.appendedFunction(Environment.onDayChanged, FogControl.onDayChanged)
Weather.onWeatherObjectChanged = Utils.appendedFunction(Weather.onWeatherObjectChanged, FogControl.onWeatherObjectChanged)

-- Override default fog randomization
Weather.randomizeFog = Utils.overwrittenFunction(Weather.randomizeFog, function(superFunc, self)
    if FogControl.fogEnabled then
        FogControl:debugPrint("Overriding default fog randomization")
        return
    end
    return superFunc(self)
end)